#!/usr/bin/env python
import requests
from lxml import etree as ET
from flask import Flask, Response
from urllib.parse import unquote
from dateutil import parser, tz
from datetime import datetime
from isodate import parse_duration
from email.utils import formatdate

from utils import read_config, hasher, Entity, Server

import logging
log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)

config = read_config('mdproxy.yaml')
app = Flask(__name__)


# Find all IdP's in edugain metadata
cached = Server()


@app.route('/<realm>/entities',
           strict_slashes=False,
           methods=['GET'])
def serve_all(realm):
    response = Response()
    response.headers['Content-Type'] = "application/samlmetadata+xml"
    response.headers['Content-Disposition'] = "filename = \"metadata.xml\""

    if cached.all_entities is not None:
        print("cache all")
        cache = cached.all_entities
        data = cache.md
        last_modified = cache.last_modified
        expires = min(datetime.now(tz.tzutc()) +
                      cache.cache_duration,
                      cache.valid_until)
        max_age = int((expires -
                       datetime.now(tz.tzutc())).total_seconds())

    else:
        print("request all")
        request = requests.get(f"{config[realm]['signer']}/{realm}"
                               f"/entities")
        data = request.text
        last_modified = request.headers.get(
            'Last-Modified',
            formatdate(timeval=None,
                       localtime=False,
                       usegmt=True)
            )

        try:
            root = ET.fromstring(data)
            validUntil = root.get('validUntil')
            cacheDuration = root.get('cacheDuration')
            cached_entity = Entity()
            cached_entity.md = data
            cached_entity.valid_until = parser.isoparse(validUntil)
            cached_entity.cache_duration = parse_duration(cacheDuration)
            cached_entity.expires = min(datetime.now(tz.tzutc())
                                        + cached_entity.cache_duration,
                                        cached_entity.valid_until)
            cached_entity.last_modified = last_modified
            max_age = int((cached_entity.expires -
                           datetime.now(tz.tzutc())).total_seconds())
            cached_entity.max_age = max_age
            if cached_entity.expires > datetime.now(tz.tzutc()):
                cached.all_entities = cached_entity
            else:
                raise KeyError
        except Exception as e:
            print(f"{e}")
            data = "No valid metadata\n"
            max_age = 60
            response.headers['Content-type'] = "text/html"
            response.headers['Cache-Control'] = "max-age=60"
            response.status = 404

    response.headers['Cache-Control'] = f"max-age={max_age}"
    response.headers['Last-Modified'] = last_modified
    response.data = data
    return response


@app.route('/<realm>/entities/<path:eid>',
           methods=['GET'])
def serve_one(realm, eid):
    entity = unquote(eid)
    if entity[:6] == "{sha1}":
        sha1 = entity[6:]
    else:
        sha1 = hasher(entity)

    response = Response()
    response.headers['Content-Type'] = "application/samlmetadata+xml"
    response.headers['Content-Disposition'] = "filename = \"metadata.xml\""

    cached[realm] = cached.get(realm, {})
    if sha1 in cached[realm]:
        if cached[realm][sha1].expires > datetime.now(tz.tzutc()):
            print(f"cache {sha1}")
            max_age = int((cached[realm][sha1].expires -
                           datetime.now(tz.tzutc())).total_seconds())
            last_modified = cached[realm][sha1].last_modified
            response.headers['Cache-Control'] = f"max-age={max_age}"
            response.headers['Last-Modified'] = last_modified
            response.data = cached[realm][sha1].md
            return response

    print(f"request {sha1}")
    request = requests.get(f"{config[realm]['signer']}/{realm}"
                           f"/entities/{{sha1}}{sha1}")
    data = request.text
    last_modified = request.headers.get('Last-Modified',
                                        formatdate(timeval=None,
                                                   localtime=False,
                                                   usegmt=True))
    try:
        root = ET.fromstring(data)
        validUntil = root.get('validUntil')
        cacheDuration = root.get('cacheDuration')
        cached_entity = Entity()
        cached_entity.md = data
        cached_entity.valid_until = parser.isoparse(validUntil)
        cached_entity.cache_duration = parse_duration(cacheDuration)
        cached_entity.expires = min(datetime.now(tz.tzutc())
                                    + cached_entity.cache_duration,
                                    cached_entity.valid_until)
        cached_entity.last_modified = last_modified
        if cached_entity.expires > datetime.now(tz.tzutc()):
            cached[realm][sha1] = cached_entity
            max_age = int((cached_entity.expires -
                           datetime.now(tz.tzutc())).total_seconds())
        else:
            raise KeyError
    except KeyError as e:
        print(f"  {sha1} {e}")
        data = "No valid metadata\n"
        max_age = 60
        response.headers['Content-type'] = "text/html"
        response.headers['Cache-Control'] = "max-age=60"
        response.status = 404

    response.headers['Cache-Control'] = f"max-age={max_age}"
    response.headers['Last-Modified'] = last_modified
    response.data = data
    return response


if __name__ == "__main__":
    app.run(host='127.0.0.1', port=5002, debug=False)

application = app
