import os
from lxml import etree as ET
from dateutil import parser, tz
from isodate import parse_duration, duration_isoformat
from datetime import datetime, timedelta
import hashlib
from urllib.parse import unquote
import yaml
import pyinotify
from signers import Signers
from constants import MD_NAMESPACE, NSMAP


def read_config(config):
    with open(config) as f:
        config = yaml.safe_load(f)
    return config


def hasher(entity_id):
    sha1 = hashlib.sha1()
    sha1.update(entity_id.encode())
    sha1_digest = sha1.hexdigest()
    return sha1_digest


class Entity:
    def __init__(self):
        self.md = None
        self.valid_until = 0
        self.cache_duration = 0
        self.last_modified = 0
        self.max_age = (datetime.now(tz.tzutc()) +
                        timedelta(seconds=60))


class EventProcessor(pyinotify.ProcessEvent):
    def __init__(self, realm):
        self.realm = realm

    def process_default(self, event):
        if event.maskname in ["IN_DELETE",
                              "IN_MOVED_FROM",
                              "IN_MOVED_TO"]:
            self.realm.walk_location(event.path)
        else:
            self.realm.read_metadata(event.pathname)


class Server(dict):
    def __init__(self):
        self.all_entities = None


class Realm:
    watch_list = {}
    dirty = False
    all_cache = None

    def __init__(self, location, signer):
        self.idps = {}
        self.mdfiles = {}
        self.signer = Signers(signer)

        self.md_watcher = pyinotify.WatchManager()
        self.md_events = pyinotify.ThreadedNotifier(self.md_watcher,
                                                    EventProcessor(self))
        self.md_events.start()

        self.walk_location(location)

    def walk_location(self, location):
        old_mdfiles = self.mdfiles.copy()
        location = os.path.realpath(location)
        self.md_watcher.add_watch(location, pyinotify.IN_CLOSE_WRITE
                                  | pyinotify.IN_DELETE
                                  | pyinotify.IN_MOVED_FROM
                                  | pyinotify.IN_MOVED_TO)

        files = os.listdir(location)
        for file in files:
            mdfile = os.path.realpath(os.path.join(location, file))
            if os.path.isfile(mdfile):
                old_mdfiles.pop(mdfile, None)
                self.read_metadata(mdfile)

        for mdf, idps in old_mdfiles.items():
            print("--- REMOVE METADATA --")
            print(mdf)
            for idp in idps:
                print(f"  {{sha1}}{idp}")
                self.idps.pop(idp, None)
                self.__dict__.pop(idp, None)

    def read_metadata(self, mdfile):
        print("--- READ METADATA --")
        print(mdfile)
        found = 0
        removed = 0
        old_idps = self.mdfiles.get(mdfile, []).copy()
        mdfiles = []
        try:
            root = ET.ElementTree(file=mdfile).getroot()
        except Exception:
            print("Invalid metadata")
            return
        ns = root.nsmap.copy()
        ns['xml'] = 'http://www.w3.org/XML/1998/namespace'
        for signature in root.findall('.//ds:Signature', ns):
            signature.getparent().remove(signature)
        validUntil = root.get('validUntil')
        cacheDuration = root.get('cacheDuration')
        valid_until = parser.isoparse(validUntil)
        cache_duration = parse_duration(cacheDuration)
        last_modified = datetime.now(tz.tzutc())
        if valid_until > datetime.now(tz.tzutc()):
            self.all_cache = None
            for entity_descriptor in root.findall('md:EntityDescriptor', ns):
                entityID = entity_descriptor.attrib.get('entityID', 'none')
                sha1 = hasher(entityID)
                print(f"  {{sha1}}{sha1} {entityID}")
                entity_descriptor.set('validUntil', validUntil)
                entity_descriptor.set('cacheDuration', cacheDuration)
                entity = Entity()
                entity.md = entity_descriptor
                entity.valid_until = valid_until
                entity.cache_duration = cache_duration
                entity.expires = min(datetime.now(tz.tzutc()) + cache_duration,
                                     valid_until)
                entity.last_modified = last_modified
                self.idps[sha1] = entity
                self.__dict__.pop(sha1, None)
                if sha1 in old_idps:
                    old_idps.remove(sha1)
                mdfiles.append(sha1)
                found += 1
        for idp in old_idps:
            self.idps.pop(idp, None)
            self.__dict__.pop(idp, None)
            removed += 1

        self.mdfiles[mdfile] = list(mdfiles)

        print(f"  Found: {found} entities")
        print(f"  Removed: {removed} entities\n")

    def __getitem__(self, key):
        entityID = unquote(key)
        if entityID[:6] == "{sha1}":
            sha1 = entityID[6:]
        else:
            sha1 = hasher(entityID)

        data = Entity()
        if sha1 in self.__dict__:
            signed_entity = self.__dict__[sha1]
            if signed_entity.expires > datetime.now(tz.tzutc()):
                print(f"cache {sha1}")
                data.md = self.__dict__[sha1].md

        if data.md is None and sha1 in self.idps:
            try:
                print(f"sign {sha1}")
                valid_until = self.idps[sha1].valid_until
                if valid_until > datetime.now(tz.tzutc()):
                    signed_element = self.signer.sign(self.idps[sha1].md)
                    signed_xml = ET.tostring(signed_element,
                                             pretty_print=True).decode()
                    signed_entity = Entity()
                    signed_entity.md = signed_xml
                    signed_entity.expires = (datetime.now(tz.tzutc()) +
                                             self.idps[sha1].cache_duration)
                    signed_entity.last_modified = self.idps[sha1].last_modified
                    self.__dict__[sha1] = signed_entity
                    data.md = signed_xml
                else:
                    raise KeyError
            except KeyError as e:
                print(f"  {sha1} {e}")

        data.max_age = int((signed_entity.expires -
                           datetime.now(tz.tzutc())).total_seconds())
        data.last_modified = signed_entity.last_modified
        return data

    def all_entities(self):
        if self.all_cache is not None:
            print("cache all")
            data = self.all_cache
        else:
            print("sign all")
            data = Entity()
            ns = NSMAP
            root = ET.Element(f"{{{MD_NAMESPACE}}}EntitiesDescriptor",
                              nsmap=ns)
            # We are going to minimize expires, so set to some inf value
            valid_until = (datetime.now(tz.tzutc()) +
                           timedelta(days=365))
            cache_duration = parse_duration("P1D")
            for sha1, entity in self.idps.items():
                valid_until = min(valid_until, entity.valid_until)
                cache_duration = min(cache_duration, entity.cache_duration)
                ET.strip_attributes(entity.md, 'validUntil', 'cacheDuration')
                root.append(entity.md)

            vu_zulu = str(valid_until).replace('+00:00', 'Z')
            root.set('validUntil', vu_zulu)
            root.set('cacheDuration', duration_isoformat(cache_duration))
            last_modified = datetime.now(tz.tzutc())

            signed_root = self.signer.sign(root)
            data.md = ET.tostring(signed_root, pretty_print=True)
            data.valid_until = valid_until
            data.last_modified = last_modified

            self.all_cache = data

        return data
